use std::convert::TryInto;
use std::path::{Component, Path, PathBuf};
use std::sync::{self, Arc, Mutex};

use rocket::{
    fs::{self, FileServer},
    get,
    http::{uri::Origin, Method, Status},
    outcome::{
        try_outcome,
        Outcome::{Failure, Success},
    },
    request::{FromRequest, Outcome, Request},
    response::{content, Response},
    routes, Route, State,
};
use thiserror::Error;
use tokio::{
    fs::File,
    io::{AsyncRead, AsyncReadExt},
};
use url::Host;
use uuid::Uuid;

use crate::Config;

pub fn routes() -> Vec<Route> {
    routes![static_websites]
}

/// Contains error for websites
#[derive(PartialEq, Clone, Debug, Error)]
pub enum Error {
    #[error("Can't lock an underling mutex")]
    Lock,
    #[error("Malformed subdomain, did you put '.' in it ?")]
    MalformedSubdomain,
    #[error("IP host are not supported")]
    IPHost,
    #[error("Internal error")]
    InternalError,
}

type Result<T, E = Error> = core::result::Result<T, E>;

/// Contains information to serve files for a static website.
///
/// A static website is identified by its host
#[derive(Debug, Clone, PartialEq)]
pub struct Website {
    /// Path of storage for the website
    path: PathBuf,
    /// Host (domain) of the website
    host: Host,
}

fn host_from_subdomain<S: AsRef<str>>(subdomain: S, base_host: &Host) -> Result<Host> {
    if subdomain.as_ref().contains(".") {
        return Err(Error::MalformedSubdomain);
    }
    let host = match base_host {
        Host::Domain(base_host) => format!("{}.{}", subdomain.as_ref(), base_host.to_string()),
        _ => return Err(Error::IPHost),
    };
    Ok(Host::parse(&host).map_err(|_| Error::MalformedSubdomain)?)
}

impl Website {
    /// Creates a website from a given subdomain and a base_host
    ///
    /// # Errors
    ///
    /// If your subdomain contains dots, an Error::MalformedUrl is returned
    /// Error::IPHost is returned if the given host is an IP
    fn from_subdomain<S: AsRef<str>>(subdomain: S, base_host: &Host) -> Result<Self> {
        // TODO: implement the path creation
        let path = Uuid::new_v4();
        let path = path.to_string().into();
        let host = host_from_subdomain(subdomain, base_host)?;
        Ok(Self { path, host })
    }

    /// Returns the file for a path if the file exists, and None if there is no file found
    async fn file<P>(&self, path: P, config: &Config) -> Option<File>
    where
        P: AsRef<Path>,
    {
        let complete_path = config.data_dir.join(self.path.join(path.as_ref()));
        let complete_path = match complete_path.is_dir() {
            // TODO: add better sanitization
            _ if complete_path.to_str().unwrap_or_default().contains("..") => None,
            true => Some(complete_path.join(Path::new("index.html"))),
            false => Some(complete_path),
        };
        match complete_path {
            None => None,
            Some(path) => File::open(path).await.ok(),
        }
    }
}

#[derive(Debug, Clone)]
pub struct Websites {
    hosts: Arc<Mutex<Vec<Website>>>,
    // TODO: Move to domain.rs
    base_host: Host,
}

impl Websites {
    pub fn new(base_host: Host) -> Self {
        Self {
            hosts: Arc::new(Mutex::new(vec![])),
            base_host,
        }
    }

    /// Returns the website indentified by the host
    ///
    /// # Errors
    ///
    /// If the data can't be accessed, returns Error::Lock
    pub fn website(&self, host: &Host) -> Result<Option<Website>> {
        match self.hosts.lock() {
            Ok(hosts) => Ok(hosts
                .iter()
                .find(|website| &website.host == host)
                .map(Clone::clone)),
            Err(_) => Err(Error::Lock),
        }
    }

    /// Returns the path containing the data of the the host if the host exists,
    /// Returns None if the subdomain is unknown
    ///
    /// # Errors
    ///
    /// If the data can't be accessed, returns Error::Lock
    pub fn host_path(&self, host: &Host) -> Result<Option<PathBuf>> {
        Ok(self.website(host)?.map(|w| w.path))
    }

    /// Returns the path containing the data of the the subdomain if the subdomain exists,
    /// Returns None if the subdomain is unknown
    ///
    /// # Errors
    ///
    /// If the data can't be accessed, returns Error::Lock
    pub fn subdomain_path<S: AsRef<str>>(&self, subdomain: S) -> Result<Option<PathBuf>> {
        self.host_path(&host_from_subdomain(subdomain, &self.base_host)?)
    }

    /// Adds a subdomain to the list of hosts
    ///
    /// Returns the PathBuf
    pub fn add_subdomain<S: AsRef<str>>(&self, subdomain: S) -> Result<PathBuf> {
        match &mut self.hosts.lock() {
            Ok(hosts) => {
                hosts.push(Website::from_subdomain(&subdomain, &self.base_host)?);
                Ok(())
            }
            Err(_) => Err(Error::Lock),
        }?;
        Ok(self
            .subdomain_path(&subdomain)?
            .ok_or(Error::InternalError)?)
    }
}

#[rocket::async_trait]
impl<'a> FromRequest<'a> for Website {
    type Error = ();

    async fn from_request(req: &'a Request<'_>) -> Outcome<Self, Self::Error> {
        let subdomain_hosts = try_outcome!(req.guard::<&State<Websites>>().await);
        match (req.headers().get_one("Host"), &subdomain_hosts) {
            // TODO: those unwrap should return Outcome Errors
            (Some(h), sh) => match sh.website(&Host::parse(h).unwrap()).unwrap() {
                None => Outcome::Forward(()),
                Some(website) => Outcome::Success(website),
            },
            (_, _) => Outcome::Forward(()),
        }
    }
}

#[get("/<path..>")]
async fn static_websites(
    path: PathBuf,
    website: Website,
    config: &State<Config>,
) -> Option<content::Html<String>> {
    // TODO: serve any file, change output format
    // TODO: switch to custom error to enhance error logging
    let mut buffer = Vec::new();
    let mut file = match website.file(&path, &config).await {
        None => return None,
        Some(f) => f,
    };
    match file.read_to_end(&mut buffer).await.ok() {
        None => None,
        Some(_) => String::from_utf8(buffer).ok().map(content::Html),
    }
}
